import unittest

from .defs import MYSQL_CONN_STR, ORACLE_CONN_STR, POSTGRESQL_CONN_STR, SQL_SERVER_CONN_STR, SQLITE_CONN_STR, FakeEngine


class TestEngine(unittest.TestCase):
    def setUp(self) -> None:
        self.postgresql_engine = FakeEngine(POSTGRESQL_CONN_STR)
        self.mysql_engine = FakeEngine(MYSQL_CONN_STR)
        self.sqlite_engine = FakeEngine(SQLITE_CONN_STR)
        self.sql_server_engine = FakeEngine(SQL_SERVER_CONN_STR)
        self.oracle_engine = FakeEngine(ORACLE_CONN_STR)
        self.no_dialect_engine = FakeEngine("yeeee://haaaaa")

    def test_postgresql_conn_str(self) -> None:
        self.assertEqual(self.postgresql_engine.dialect, "postgresql")

    def test_mysql_conn_str(self) -> None:
        self.assertEqual(self.mysql_engine.dialect, "mysql")

    def test_sqlite_conn_str(self) -> None:
        self.assertEqual(self.sqlite_engine.dialect, "sqlite")

    def test_sql_server_conn_str(self) -> None:
        self.assertEqual(self.sql_server_engine.dialect, "sql_server")

    def test_oracle_conn_str(self) -> None:
        self.assertEqual(self.oracle_engine.dialect, "oracle")

    def test_no_dialect_conn_str(self) -> None:
        self.assertEqual(self.no_dialect_engine.dialect, "yeeee")

    def test_no_schema_conn_str(self) -> None:
        with self.assertRaises(ValueError):
            _ = FakeEngine("djksjjkjdskdjdj")



"""Contains definitions for frequently used composite types."""

from typing import Type, TypeAlias

from .column import BaseColumn
from .subquery import Subquery

QueryEntity: TypeAlias = Type | BaseColumn
BaseColumnRef: TypeAlias = BaseColumn | str
JoinEntity: TypeAlias = Type | Subquery



"""Contains basic datatypes."""

from datetime import date, datetime, time
from decimal import Decimal

from ..exceptions import ValueOutOfBound


class IntegerMeta(type):
    @property
    def __name__(cls) -> str:
        return "Integer"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, int)


class Integer(metaclass=IntegerMeta):
    pass


class BooleanMeta(type):
    @property
    def __name__(cls) -> str:
        return "Boolean"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, bool)


class Boolean(metaclass=BooleanMeta):
    pass


class FloatMeta(type):
    @property
    def __name__(cls) -> str:
        return "Float"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, float)


class Float(metaclass=FloatMeta):
    pass


class NumericMeta(type):
    @property
    def __name__(cls) -> str:
        return "Numeric"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, Decimal)


class Numeric(metaclass=NumericMeta):
    pass


class StringMeta(type):
    @property
    def __name__(cls) -> str:
        return "String"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, str)


class String(metaclass=StringMeta):
    pass


class DateMeta(type):
    @property
    def __name__(cls) -> str:
        return "Date"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, date)


class Date(metaclass=DateMeta):
    pass


class TimeMeta(type):
    @property
    def __name__(cls) -> str:
        return "Time"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, time)


class Time(metaclass=TimeMeta):
    pass


class DateTimeMeta(type):
    @property
    def __name__(cls) -> str:
        return "DateTime"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, datetime)


class DateTime(metaclass=DateTimeMeta):
    pass


class BinaryMeta(type):
    @property
    def __name__(cls) -> str:
        return "Binary"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, bytes)


class Binary(metaclass=BinaryMeta):
    pass


class Serial(Integer):
    """ORM representation of serial types in SQL variants. Inherits from :class:`.backends.Integer`.

    :params val:    underlying integer value.
    """

    def __init__(self, val: Integer) -> None:
        self.val = val
        if self.val < 1 or self.val > 2147483647:
            raise ValueOutOfBound("Serial")



"""Contains :class:`.subquery.Subquery` and its helpers."""

from typing import Any, List

from .column import VirtualField
from .exceptions import EntityError


class Subquery:
    """Representer of an SQL subquery.

    A subquery is a nested SELECT statement that is used within another SQL statement.

    Never directly instantiated, but rather initialised by invoking :meth:`.query.Query.subquery()`.

    :param inner_sql:       SQL string of the view represented by the subquery
    :param column_names:    original names of the columns in that view.
    """

    subquery_idx = 0

    def __init__(self, inner_sql: str, column_names: List[str]) -> None:
        type(self).subquery_idx += 1
        self.inner_sql = inner_sql
        self.column_names = column_names
        self.view_name = "anon_%d" % self.subquery_idx
        self.full_sql = "(%s) AS %s" % (self.inner_sql, self.view_name)

    def __str__(self) -> str:
        return self.inner_sql

    def __getattr__(self, attr: str) -> Any:
        if attr not in self.column_names:
            raise EntityError("Wrong field for subquery.")

        return VirtualField(attr, self.view_name)



"""Contains library-wide constants."""

TYPES = {
    "postgresql": {
        "Integer": "INTEGER",
        "Boolean": "BOOLEAN",
        "Float": "FLOAT",
        "Numeric": "NUMERIC",
        "String": "VARCHAR",
        "Date": "DATE",
        "Time": "TIME",
        "DateTime": "TIMESTAMP",
        "Binary": "BYTEA",
        "Geometry": "GEOMETRY",
        "Interval": "INTERVAL",
        "Serial": "SERIAL",
        "ARRAY": "ARRAY",
        "JSONB": "JSONB",
        "CIDR": "CIDR",
        "MACADDR": "MACADDR",
        "HSTORE": "HSTORE",
    },
    "mysql": {
        "Integer": "INT",
        "Boolean": "TINYINT(1)",
        "Float": "FLOAT",
        "Numeric": "DECIMAL",
        "String": "VARCHAR",
        "Date": "DATE",
        "Time": "TIME",
        "DateTime": "DATETIME",
        "Binary": "BLOB",
        "Geometry": "GEOMETRY",
    },
    "sqlite": {
        "Integer": "INTEGER",
        "Boolean": "INTEGER",
        "Float": "REAL",
        "Numeric": "NUMERIC",
        "String": "TEXT",
        "Date": "DATE",
        "Time": "TIME",
        "DateTime": "DATETIME",
        "Binary": "BLOB",
    },
    "sql_server": {
        "Integer": "INT",
        "Boolean": "BIT",
        "Float": "FLOAT",
        "Numeric": "DECIMAL",
        "String": "VARCHAR",
        "Date": "DATE",
        "Time": "TIME",
        "DateTime": "DATETIME",
        "Binary": "VARBINARY",
        "Geometry": "GEOMETRY",
    },
    "oracle": {
        "Integer": "NUMBER",
        "Boolean": "NUMBER(1)",
        "Float": "NUMBER",
        "Numeric": "NUMBER",
        "String": "VARCHAR2",
        "Date": "DATE",
        "Time": "TIMESTAMP",
        "DateTime": "TIMESTAMP",
        "Binary": "BLOB",
        "Geometry": "SDO_GEOMETRY",
        "Interval": "INTERVAL",
    },
}



"""Contains datatypes to support PostgreSQL backend."""

import ipaddress
import json
from datetime import timedelta

import macaddress
from shapely.geometry.base import BaseGeometry


class GeometryMeta(type):
    @property
    def __name__(cls) -> str:
        return "Geometry"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, BaseGeometry)


class Geometry(metaclass=GeometryMeta):
    pass


class IntervalMeta(type):
    @property
    def __name__(cls) -> str:
        return "Interval"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, timedelta)


class Interval(metaclass=IntervalMeta):
    pass


class ARRAYMeta(type):
    @property
    def __name__(cls) -> str:
        return "ARRAY"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, list)


class ARRAY(metaclass=ARRAYMeta):
    pass


class JSONBMeta(type):
    @property
    def __name__(cls) -> str:
        return "JSONB"

    def __instancecheck__(cls, instance) -> bool:
        try:
            _ = json.loads(instance)
        except json.JSONDecodeError:
            return False
        return True


class JSONB(metaclass=JSONBMeta):
    pass


class CIDRMeta(type):
    @property
    def __name__(cls) -> str:
        return "CIDR"

    def __instancecheck__(self, instance) -> bool:
        return isinstance(instance, (ipaddress.IPv4Network, ipaddress.IPv6Network))


class CIDR(metaclass=CIDRMeta):
    pass


class MACADDRMeta(type):
    @property
    def __name__(cls) -> str:
        return "MACADDR"

    def __instancecheck__(self, instance) -> bool:
        return isinstance(instance, macaddress.MAC)


class MACADDR(metaclass=MACADDRMeta):
    pass


class HSTOREMeta(type):
    @property
    def __name__(cls) -> str:
        return "HSTORE"

    def __instancecheck__(self, instance) -> bool:
        return isinstance(instance, dict)


class HSTORE(metaclass=HSTOREMeta):
    pass



"""Contains datatypes to support Oracle DB backend."""

from datetime import timedelta

from shapely.geometry.base import BaseGeometry


class GeometryMeta(type):
    @property
    def __name__(cls) -> str:
        return "Geometry"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, BaseGeometry)


class Geometry(metaclass=GeometryMeta):
    pass


class IntervalMeta(type):
    @property
    def __name__(cls) -> str:
        return "Interval"

    def __instancecheck__(cls, instance) -> bool:
        return isinstance(instance, timedelta)


class Interval(metaclass=IntervalMeta):
    pass



"""Contains abstract and concrete database engine classes, as well as the dialect inferrer class."""

from urllib.parse import urlparse

import pyodbc

from .pool import ConnectionPool


class DialectInferrer:
    """Delegatee class concerning with encapsulation of the sql dialect inferrence functionality.

    :param conn_str:    The database connection string used to infer the SQL dialect.
    """

    def __init__(self, conn_str: str) -> None:
        self.conn_str = conn_str

    @property
    def sql_dialect(self) -> str:
        """Infers the SQL dialect from the connection string."""
        parsed_url = urlparse(self.conn_str)
        if not parsed_url.scheme:
            raise ValueError("Connection string is missing a scheme.")

        dialect = parsed_url.scheme
        if "+" in dialect:
            dialect = dialect.split("+")[0]

        if dialect in ("mssql", "sqlserver"):
            dialect = "sql_server"

        return dialect


class AbstractEngine:
    """Abstract database engine class."""

    active_instance = None


class DBEngine(AbstractEngine):
    """Connection adapter for the :class:`.db_session.DBSession` object.

    Downstream from a thread-safe connection pool. This connection pool uses lazy loading.

    The class keeps record of the most recent active instance as an inner state.

    :param conn_str:    database location, along with auth params
    :param pool_size:   keyword-only. Size of the connection pool.
    """

    def __init__(self, conn_str: str, *, pool_size: int) -> None:
        self.dialect_inferrer = DialectInferrer(conn_str)
        self.dialect = self.dialect_inferrer.sql_dialect
        self.connection_pool = ConnectionPool(conn_str, pool_size)
        self.conn = self.get_connection()
        self.cursor = self.conn.cursor()

    def get_connection(self) -> pyodbc.Connection:
        """Gets a connection from the pool."""
        return self.connection_pool.acquire()

    def release_connection(self, conn: pyodbc.Connection) -> None:
        """Releases a connection back to the pool."""
        self.connection_pool.release(conn)
